#include "KMUninstall.h"

#include <memory>

#include "common/KMError.h"
#include "common/KMFolder.h"
#include "common/KMLogger.h"
#include "common/KMMetaData.h"
#include "common/KMOSTreeHandler.h"

REGISTER_SUBCOMMAND_DYNCREATE(uninstall, KMUninstall)

class KMUninstall::Private
{
public:
    shared_ptr<KMFolder> m_folder;
    map<string, KMRef> m_mapRefs;
    shared_ptr<KMOSTreeHandler> m_ostreeHandler;
};

KMUninstall::KMUninstall()
    : m_dPtr(make_shared<Private>())
{
    m_dPtr->m_folder = make_shared<KMFolder>(KMFolder());
    m_dPtr->m_ostreeHandler = shared_ptr<KMOSTreeHandler>(new KMOSTreeHandler(m_dPtr->m_folder->getRepoPath()));
    m_dPtr->m_mapRefs = m_dPtr->m_ostreeHandler->getLocalRefs(BASE_REOMTE);
}

KMUninstall::~KMUninstall() { }

int KMUninstall::despose(int argc, char **argv)
{
    vector<string> vecUninstall;

    for (int i = 1; i < argc; i++)
    {
        if (argv[i][0] != '-')
        {
            vecUninstall.push_back(argv[i]);
        }
        else if (vecUninstall.size() > 0)
        {
            kmlogger.error("args error");
            return 1;
        }
    }
    return process(vecUninstall);
}

/// @brief 暂时只支持卸载 app，runtime 不支持卸载
/// @param vec_name 要卸载的 app
/// @return
int KMUninstall::process(const vector<string> &vec_name)
{
    for (auto name : vec_name)
    {
        if (!m_dPtr->m_folder->isInstalled(name))
        {
            kmlogger.pinfo("%s not installed!", name.c_str());
            continue;
        }
        kmlogger.pinfo("find installed package %s", name.c_str());

        // 通过元数据文件解析需要删除的包
        unique_ptr<KMMetaData> metadata;
        if (m_dPtr->m_folder->isApp(name))
        {
            metadata = unique_ptr<KMMetaData>(new KMMetaData());

            if (!metadata->load(m_dPtr->m_folder->getActivePath(name) / "metadata"))
            {
                continue;
            }

            fs::path exportPath = m_dPtr->m_folder->getExportsPath();
            if (std::filesystem::exists(exportPath))
            {
                for (auto const &entryDir : fs::recursive_directory_iterator(exportPath))
                {
                    if (fs::is_symlink(entryDir) && (int)fs::read_symlink(entryDir).string().find(name) > 0)
                    {
                        kmlogger.pinfo("remove %s", entryDir.path().c_str());
                        if (!fs::remove_all(entryDir))
                        {
                            kmlogger.perror("uninstall %s error, can\'t delete exports file.", name.c_str());
                            return KM_ERROR_DELETE_EXPORT_FILES_FAILED;
                        }
                    }
                }
            }

            fs::path varExportPath = m_dPtr->m_folder->getVarExportsPath();

            std::vector<std::string> subDirs = { "share", "bin" };
            for (const auto &subDir : subDirs)
            {
                fs::path varExportSubDir = varExportPath / subDir;

                if (std::filesystem::exists(varExportSubDir))
                {
                    for (auto const &entryDir : fs::recursive_directory_iterator(varExportSubDir))
                    {
                        if (fs::is_symlink(entryDir) && (int)fs::read_symlink(entryDir).string().find(name) > 0)
                        {
                            kmlogger.pinfo("remove %s", entryDir.path().c_str());
                            if (!fs::remove_all(entryDir))
                            {
                                kmlogger.perror("uninstall %s error, can\'t delete exports file.", name.c_str());
                                return KM_ERROR_DELETE_EXPORT_FILES_FAILED;
                            }
                        }
                    }
                }
            }

            // 删除文件夹
            kmlogger.pinfo("remove app %s install dir.", name.c_str());
            if (!fs::remove_all(m_dPtr->m_folder->getAppPath() / name))
            {
                kmlogger.error("uninstall %s error, can\'t delete the directory.", name);
                return KM_ERROR_DELETE_DIRECTORY_FAILED;
            }
        }
        else if (m_dPtr->m_folder->isRuntime(name))
        {
            metadata = unique_ptr<KMMetaData>(new KMMetaData(false));
            // 卸载运行时要考虑有没有应用依赖此运行时

            for (auto pair : m_dPtr->m_mapRefs)
            {
                if (pair.second.id == name)
                {
                    metadata->load(m_dPtr->m_folder->getRuntimePath() / name / pair.second.arch / pair.second.branch / "active" / "metadata");
                    break;
                }
            }
            // 删除文件夹
            kmlogger.info("remove runtime %s install dir", name.c_str());
            if (!fs::remove_all(m_dPtr->m_folder->getRuntimePath() / name))
            {
                kmlogger.error("uninstall %s error, can\'t delete the directory.", name);
                return KM_ERROR_DELETE_DIRECTORY_FAILED;
            }
        }

        // 从本地 repo 中删除
        for (auto pair : m_dPtr->m_mapRefs)
        {
            if (pair.second.id == name)
            {
                string refString = pair.second.kind + "/" + pair.second.id + "/" + pair.second.arch + "/" + pair.second.branch;

                if (!m_dPtr->m_ostreeHandler->deleteRefTag("", string("deploy/").append(refString)))
                {
                    return KM_ERROR_DELETE_DEPLOY_TAG_FAILED;
                }

                if (!m_dPtr->m_ostreeHandler->deleteRefTag(BASE_REOMTE, refString))
                {
                    return KM_ERROR_DELETE_REF_TAG_FAILED;
                }

                if (!m_dPtr->m_ostreeHandler->repoPrune())
                {
                    return KM_ERROR_PRUNE_REPO_FAILED;
                }
            }
        }
        kmlogger.pinfo("\033[1;32muninstall %s ok.\033[0m", name.c_str());

        // 删除扩展
        for (auto ext : metadata->getExtensions())
        {
            kmlogger.info("remove extension %s", ext.name.c_str());
            if (ext.autodelete && m_dPtr->m_folder->isInstalled(ext.name))
            {
                int ret = process(vector<string>{ ext.name });
                if (ret != KM_ERROR_NO)
                {
                    return ret;
                }
            }
        }
    }
    return KM_ERROR_NO;
}

vector<string> KMUninstall::getRuntimeApp(const string &name)
{
    // 遍历 app 查看依赖

    return vector<string>();
}
